@@ -46,8 +46,9 @@ func createAndPutKeys(t *testing.T) {
 			}
 
 			var key [16]byte
-			if _, err := rand.Read(key[:]); err != nil {
-				return err
+			_, rerr := rand.Read(key[:])
+			if rerr != nil {
+				return rerr
 			}
 			if err := nodes.Put(key[:], nil); err != nil {
 				return err
